package org.apache.commons.lang3;

import java.util.HashMap;
import java.io.ObjectStreamClass;
import java.util.Map;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.Serializable;

public class SerializationUtils
{
    public static <T extends Serializable> T clone(final T object) {
        if (object == null) {
            return null;
        }
        final byte[] objectData = serialize(object);
        final ByteArrayInputStream bais = new ByteArrayInputStream(objectData);
        try (ClassLoaderAwareObjectInputStream in = new ClassLoaderAwareObjectInputStream(bais, object.getClass().getClassLoader())) {
            final T readObject = (T)in.readObject();
            return readObject;
        }
        catch (final ClassNotFoundException ex) {
            throw new SerializationException("ClassNotFoundException while reading cloned object data", ex);
        }
        catch (final IOException ex2) {
            throw new SerializationException("IOException while reading or closing cloned object data", ex2);
        }
    }
    
    public static <T> T deserialize(final byte[] objectData) {
        Validate.notNull(objectData, "objectData", new Object[0]);
        return deserialize(new ByteArrayInputStream(objectData));
    }
    
    public static <T> T deserialize(final InputStream inputStream) {
        Validate.notNull(inputStream, "inputStream", new Object[0]);
        try (ObjectInputStream in = new ObjectInputStream(inputStream)) {
            final T obj = (T)in.readObject();
            return obj;
        }
        catch (final ClassNotFoundException | IOException ex) {
            throw new SerializationException(ex);
        }
    }
    
    public static <T extends Serializable> T roundtrip(final T obj) {
        return deserialize(serialize(obj));
    }
    
    public static byte[] serialize(final Serializable obj) {
        final ByteArrayOutputStream baos = new ByteArrayOutputStream(512);
        serialize(obj, baos);
        return baos.toByteArray();
    }
    
    public static void serialize(final Serializable obj, final OutputStream outputStream) {
        Validate.notNull(outputStream, "outputStream", new Object[0]);
        try (ObjectOutputStream out = new ObjectOutputStream(outputStream)) {
            out.writeObject(obj);
        }
        catch (final IOException ex) {
            throw new SerializationException(ex);
        }
    }
    
    static class ClassLoaderAwareObjectInputStream extends ObjectInputStream
    {
        private static final Map<String, Class<?>> primitiveTypes;
        private final ClassLoader classLoader;
        
        ClassLoaderAwareObjectInputStream(final InputStream in, final ClassLoader classLoader) throws IOException {
            super(in);
            this.classLoader = classLoader;
        }
        
        @Override
        protected Class<?> resolveClass(final ObjectStreamClass desc) throws IOException, ClassNotFoundException {
            final String name = desc.getName();
            try {
                return Class.forName(name, false, this.classLoader);
            }
            catch (final ClassNotFoundException ex) {
                try {
                    return Class.forName(name, false, Thread.currentThread().getContextClassLoader());
                }
                catch (final ClassNotFoundException cnfe) {
                    final Class<?> cls = ClassLoaderAwareObjectInputStream.primitiveTypes.get(name);
                    if (cls != null) {
                        return cls;
                    }
                    throw cnfe;
                }
            }
        }
        
        static {
            (primitiveTypes = new HashMap<String, Class<?>>()).put("byte", Byte.TYPE);
            ClassLoaderAwareObjectInputStream.primitiveTypes.put("short", Short.TYPE);
            ClassLoaderAwareObjectInputStream.primitiveTypes.put("int", Integer.TYPE);
            ClassLoaderAwareObjectInputStream.primitiveTypes.put("long", Long.TYPE);
            ClassLoaderAwareObjectInputStream.primitiveTypes.put("float", Float.TYPE);
            ClassLoaderAwareObjectInputStream.primitiveTypes.put("double", Double.TYPE);
            ClassLoaderAwareObjectInputStream.primitiveTypes.put("boolean", Boolean.TYPE);
            ClassLoaderAwareObjectInputStream.primitiveTypes.put("char", Character.TYPE);
            ClassLoaderAwareObjectInputStream.primitiveTypes.put("void", Void.TYPE);
        }
    }
}
